#!/usr/bin/env python
"""test program to exercise multicast send/receive."""

from __future__ import print_function


import sys, time
from socket import socket, AF_INET, AF_INET6, SOCK_DGRAM, IPPROTO_UDP,          \
    SOL_SOCKET, SO_REUSEADDR, IPPROTO_IP, IPPROTO_IPV6, IP_ADD_MEMBERSHIP,          \
    IPV6_JOIN_GROUP, IP_MULTICAST_TTL, IPV6_MULTICAST_HOPS, IP_MULTICAST_LOOP,          \
    IPV6_MULTICAST_LOOP, inet_pton
import socket as net
from struct import pack

from golang import b, u
from golang import sync, context


group4 = '224.1.1.1'
group6 = 'ff1e:7079:7468:6f6e:6465:6d6f:6d63:6173'
port   =  5678

def _udpsockfor(ipaddr): # -> (sk, isipv6)
    ip6 = (':' in ipaddr)  # autodetect ip4/ip6 based on group addr
    if ip6:
        sk = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP)
    else:
        sk = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)
    return (sk, ip6)


# mjoin_tx returns socket prepated to send/receive to/from multicast group:port.
def mjoin_tx(group, port, ttl=2):
    sk, ip6 = _udpsockfor(group)

    # set ttl and default destination address (for tx)
    if ip6:
        sk.setsockopt(IPPROTO_IPV6, IPV6_MULTICAST_HOPS, ttl)
    else:
        sk.setsockopt(IPPROTO_IP, IP_MULTICAST_TTL, ttl)

    sk.connect((group, port))

    # so that what we send is received locally as well
    if ip6:
        sk.setsockopt(IPPROTO_IPV6, IPV6_MULTICAST_LOOP, 1)
    else:
        sk.setsockopt(IPPROTO_IP, IP_MULTICAST_LOOP, 1)

    return sk

# mjoin_rx returns socket prepated to send/receive to/from multicast group:port.
def mjoin_rx(group, port):
    sk, ip6 = _udpsockfor(group)

    sk.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    sk.bind((group, port))  # works for reception even from multiple ifaces
    #sk.bind(('', port))     # XXX and later IP_MULTICAST_ALL=0 ?

    # join the group on all interfaces
    for ifidx, ifname in net.if_nameindex():
        print("joining %s @ [%d]%s" % (group, ifidx, ifname))
        bifidx = pack("@i", ifidx)
        if ip6:
            mreq = inet_pton(AF_INET6, group) + \
                   bifidx
            sk.setsockopt(IPPROTO_IPV6, IPV6_JOIN_GROUP, mreq)  # = IPV6_ADD_MEMBERSHIP
        else:
            mreq = inet_pton(AF_INET, group)     + \
                   inet_pton(AF_INET, '0.0.0.0') + \
                   bifidx
            #print(repr(mreq), len(mreq))
            sk.setsockopt(IPPROTO_IP, IP_ADD_MEMBERSHIP, mreq)

    return sk


def txloop(ctx, sk, pkt):
    i = 0
    while 1:
        if ctx.err():
            raise ctx.err()
        i += 1
        pkt_ = pkt + (b'.%d' % i)
        print("tx: %s ..." % u(pkt_))
        sk.send(pkt_)
        time.sleep(1)

def rxloop(ctx, sk):
    while 1:
        if ctx.err():
            raise ctx.err()
        pkt = sk.recv(4096)
        print("rx: %s" % u(pkt))


def main():
    action = sys.argv[1]

    G4 = (group4, port)
    G6 = (group6, port)
    G = None
    if "4" in action:
        G = G4
    if "6" in action:
        G = G6
    if G is None:
        raise RuntimeError("4/6 must be specified")

    wg = sync.WorkGroup(context.background())
    if "tx" in action:
        sktx = mjoin_tx(*G)
        wg.go(txloop, sktx, b(sys.argv[2]))
    if "rx" in action:
        skrx = mjoin_rx(*G)
        wg.go(rxloop, skrx)

    wg.wait()

if __name__ == '__main__':
    main()
